// iPaq 3D engine  -  Janurary 2001  -  The Phantom
// Minimal 3D engine for Handhelds
// ------------------------------------------------

#ifndef ENGINE_H
#define ENGINE_H

#define PIXEL unsigned short
#define FPP 10
#define FP_ONE (1<<FPP)
#define FPMUL(x,y) (((x)*(y))>>FPP)
#define FPDIV(x,y) (((x)<<FPP)/(y))
#define FP(x) ((int)((x)*FP_ONE+0.5f))
#define IFP(x) ((x)<<FPP)
#define PI 3.14159265359

// ------------------------------------------------
// class AVector
// Fixed point vector class
// ------------------------------------------------
class AVector
{
public:
	AVector() { m_Cell[0] = m_Cell[1] = m_Cell[2] = 0; }
	AVector( int a_X, int a_Y, int a_Z ) { m_Cell[0] = a_X; m_Cell[1] = a_Y; m_Cell[2] = a_Z; }
	~AVector() {};
	void setxyz( int a_X, int a_Y, int a_Z ) { m_Cell[0] = a_X; m_Cell[1] = a_Y; m_Cell[2] = a_Z; }
	void setx( int a_X ) { m_Cell[0] = a_X; }
	void sety( int a_Y ) { m_Cell[1] = a_Y; }
	void setz( int a_Z ) { m_Cell[2] = a_Z; }
	void set( int n, int a_Val ) { m_Cell[n] = a_Val; }
	int getx( void ) { return m_Cell[0]; }
	int gety( void ) { return m_Cell[1]; }
	int getz( void ) { return m_Cell[2]; }
	int get( int n ) { return m_Cell[n]; }
	int dot( AVector a_Second );
	AVector cross( AVector a_Second );
	int length( void );
	void normalize( void );
private:
	int m_Cell[3];
};

// ------------------------------------------------
// class AMatrix
// Fixed point matrix class
// ------------------------------------------------
class AMatrix
{
public:
	AMatrix() { identity(); }
	~AMatrix() {};
	void identity();
	void rotatex( int a_X );
	void rotatey( int a_Y );
	void rotatez( int a_Z );
	void translate( int a_X, int a_Y, int a_Z );
	void concat( const AMatrix& a_M2);
	void rotate( int a_RX, int a_RY, int a_RZ );
	int get( int n ) { return m_Cell[n]; }
	AVector getz();
private:
	int m_Cell[16];
};

// ------------------------------------------------
// class ACoord
// 3D coordinate class (original, rotated and screen)
// ------------------------------------------------
class ACoord
{
public:
	ACoord() { m_Orig.setxyz( 0, 0, 0 ); }
	ACoord( int a_X, int a_Y, int a_Z ) { m_Orig.setxyz( a_X, a_Y, a_Z ); }
	ACoord( AVector a_V ) { m_Orig = a_V; }
	~ACoord() {};
	void setorig( AVector a_V ) { m_Orig = a_V; }
	void setrot( AVector a_V ) { m_Rot = a_V; }
	AVector getorig( void ) { return m_Orig; }
	AVector getrot( void ) { return m_Rot; }
	void setscreenx( int a_X ) { m_SX = a_X; }
	void setscreeny( int a_Y ) { m_SY = a_Y; }
	int getscreenx( void ) { return m_SX; }
	int getscreeny( void ) { return m_SY; }
	void transform( AMatrix a_Mat );
	void perspective( void );
private:
	AVector m_Orig, m_Rot;
	int m_SX, m_SY;
};

// ------------------------------------------------
// class AVertex
// Vertex with two U/V's (one for fake phong or env)
// ------------------------------------------------
class AVertex
{
public:
	AVertex() { m_Crd = 0; }
	AVertex( ACoord* a_Crd, int a_U, int a_V ) { m_Crd = a_Crd; m_U1 = a_U; m_V1 = a_V; }
	~AVertex() {};
	void setcoord( ACoord* a_Crd ) { m_Crd = a_Crd; }
	ACoord* getcoord( void ) { return m_Crd; }
	void setuv1( int a_U, int a_V ) { m_U1 = a_U; m_V1 = a_V; }
	void setuv2( int a_U, int a_V ) { m_U2 = a_U; m_V2 = a_V; }
	int getu1( void ) { return m_U1; }
	int getv1( void ) { return m_V1; }
	int getu2( void ) { return m_U2; }
	int getv2( void ) { return m_V2; }
	void setrgb( int a_R, int a_G, int a_B ) { m_R = a_R; m_G = a_G; m_B = a_B; }
	int getr( void ) { return m_R; }
	int getg( void ) { return m_G; }
	int getb( void ) { return m_B; }
private:
	ACoord* m_Crd;
	int m_U1, m_V1;
	int m_U2, m_V2;
	int m_R, m_G, m_B;
};

// ------------------------------------------------
// class ATexture
// Basic texture class
// ------------------------------------------------
class ATexture
{
public:
	ATexture() { m_Bitmap = 0; }
	ATexture( int a_Width, int a_Height );
#ifdef _WIN32_WCE
	ATexture( unsigned short* a_File, int a_Width, int a_Height ) { m_Bitmap = 0; loadtga( a_File, a_Width, a_Height ); }
#else
	ATexture( char* a_File, int a_Width, int a_Height ) { m_Bitmap = 0; loadtga( a_File, a_Width, a_Height ); }
#endif
	~ATexture() { delete m_Bitmap; }
	void setsize( int a_W, int a_H ) { m_Width = a_W; m_Height = a_H; }
	int getwidth( void ) { return m_Width; }
	int getheight( void ) { return m_Height; }
	void setbitmap( PIXEL* a_Bitmap ) { m_Bitmap = a_Bitmap; }
	PIXEL* getbitmap( void ) { return m_Bitmap; }
#ifdef _WIN32_WCE
	void loadtga( unsigned short* a_File, int a_Width, int a_Height );
#else
	void loadtga( char* a_File, int a_Width, int a_Height );
#endif
private:
	int m_Width, m_Height;
	PIXEL* m_Bitmap;
};

// ------------------------------------------------
// class APolygon
// Basic textured polygon class
// ------------------------------------------------
class APolygon
{
public:
	APolygon();
	APolygon( AVertex* a_V1, AVertex* a_V2, AVertex* a_V3 ) { m_Vert[0] = a_V1; m_Vert[1] = a_V2; m_Vert[2] = a_V3; }
	~APolygon();
	void setvertex( int n, AVertex* a_Vert ) { m_Vert[n] = a_Vert; }
	AVertex* getvertex( int n ) { return m_Vert[n]; }
	void settexture( ATexture* a_Text ) { m_Text = a_Text; }
	ATexture* gettexture( void ) { return m_Text; }
	void abcd();
	AVector getnormal( void ) { return AVector( m_A, m_B, m_C ); }
	int getA( void ) { return m_A; }
	int getB( void ) { return m_B; }
	int getC( void ) { return m_C; }
	bool facing();
private:
	AVertex* m_Vert[3];
	ATexture* m_Text;
	int m_A, m_B, m_C, m_D;
};

// ------------------------------------------------
// class AMesh
// Basic mesh class
// ------------------------------------------------
class ARasterizer;
class AMesh
{
public:
	AMesh();
	~AMesh();
	void init( int a_Polies, int a_Crds );
	int coords( void ) { return m_Coords; }
	int polies( void ) { return m_Polies; }
	void transform( AMatrix a_Mat );
	void draw( ARasterizer* a_Rast, AVector a_Dir );
	ACoord** getcoords( void ) { return m_Crd; }
	APolygon** getpolies( void ) { return m_Poly; }
	void addpolygon( APolygon* a_Poly ) { m_Poly[m_Polies++] = a_Poly; }
	void addcoord( ACoord* a_Crd ) { m_Crd[m_Coords++] = a_Crd; }
	long eatchunk( char* buffer );
#ifdef _WIN32_WCE
	void load3ds( unsigned short* filename );
#else
	void load3ds( char* filename );
#endif
private:
	int m_Coords, m_Polies;
	ACoord** m_Crd;
	APolygon** m_Poly;
};

#endif

// ------------------------------------------------
// EOF